#![allow(unused_imports)]

#[macro_use]
extern crate log;
extern crate simple_logger;

#[cfg(windows)]
extern crate winapi;

#[cfg(unix)]
extern crate libc;

use std::mem::size_of; // buffer/struct size
use std::ptr::null_mut; // NULL
use std::time::Duration;

use std::mem::MaybeUninit;

// https://crates.io/crates/hwloc

/*
mod windows;
#[cfg(winows)]
use windows::{get_cpu_info, set_process_affinity};
#[cfg(unix)]
mod unix;
#[cfg(unix)]
use unix::{get_cpu_info, set_process_affinity};
*/

#[cfg(winows)]
#[path = "windows.rs"]
mod plattform;

#[cfg(unix)]
#[path = "unix.rs"]
mod plattform;

use plattform::{get_cpu_info, set_process_affinity};

fn main() {
    //let mut state = State::new();
    let cpu_info = get_cpu_info();
    info!("Hello, {:?}!", cpu_info);

    let cpu_info = cpu_info.expect("unable to get CPU infos");

    // https://docs.microsoft.com/en-us/windows/win32/procthread/numa-support

    //let cores_to_use = vec![CoreId(0), CoreId(2), CoreId(4), CoreId(6)];
    let cores_to_use = vec![CoreId(0), CoreId(1), CoreId(2), CoreId(3)];
    //let cores_to_use = vec![CoreId(0), CoreId(1)];
    set_process_affinity("*", &cores_to_use, &cpu_info.all_system_cores);

    let three_secs = Duration::from_secs(3);

    /* Demo the switching
    let proc_handle = unsafe { GetCurrentProcess() };
    for i in 0..cpu_info.unwrap() {
        unsafe { SetProcessAffinityMask(proc_handle, (1 << i) as u32) };
        make_load(three_secs * 3);
    }
    */

    info!("complete");
    std::thread::sleep(three_secs);
}

fn make_load(timeout: Duration) {
    let start = std::time::Instant::now();
    let mut i = 0usize;
    while start.elapsed() < timeout {
        if i < 1024 {
            i += 1;
        } else {
            i = 0;
        }
    }
}

#[derive(Debug)]
pub struct ProcessHandle<T> {
    pid: Pid,
    handle: T,
    name: String,
}

#[derive(Debug)]
pub struct CpuInfo {
    all_system_cores: Vec<CoreId>,
}
impl CpuInfo {
    fn n_cores(&self) -> usize {
        // lpSystemAffinityMask.count_ones() as usize
        self.all_system_cores.len()
    }
}

#[derive(Debug, PartialEq)]
pub struct CoreId(u8);

#[derive(Debug, Clone, PartialEq)]
pub struct Pid(u32);

#[cfg(windows)]
impl Pid {
    //use super::{Pid, K32EnumProcesses, DWORD, size_of};

    const N_PIDS: usize = 4096;

    pub fn get_system_pids() -> Result<Vec<Pid>, &'static str> {
        let mut pids = vec![Pid(0); Self::N_PIDS];
        let mut found_processes: DWORD = 0;

        // https://docs.microsoft.com/en-us/windows/win32/api/psapi/nf-psapi-enumprocesses
        if 0 == unsafe {
            K32EnumProcesses(
                pids.as_mut_ptr() as *mut u32,
                (Self::N_PIDS * size_of::<Pid>()) as u32,
                &mut found_processes,
            )
        } {
            return Err("unable to EnumProcesses");
        }
        let n_pids = found_processes as usize / size_of::<DWORD>();

        pids.truncate(n_pids);

        Ok(pids)
    }

    pub fn process_handle(self) -> Result<ProcessHandle, &'static str> {
        use winapi::um::winnt::{
            PROCESS_QUERY_INFORMATION, PROCESS_SET_INFORMATION, PROCESS_VM_READ, PROCESS_VM_WRITE,
        };
        // https://docs.microsoft.com/en-us/windows/win32/procthread/process-security-and-access-rights
        let desired_access = PROCESS_VM_READ | PROCESS_QUERY_INFORMATION | PROCESS_SET_INFORMATION;
        let inherit_handle = 0;

        // https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocess
        let handle: HANDLE = unsafe { OpenProcess(desired_access, inherit_handle, self.0) };
        if handle.is_null() {
            return Err("unable to OpenProcess");
        }

        let mut module_ptr: HMODULE = null_mut();
        let mut module_dyn_needed: DWORD = 0;
        let filter_flag = winapi::um::psapi::LIST_MODULES_ALL;
        if 0 == unsafe {
            K32EnumProcessModulesEx(
                handle,
                &mut module_ptr as *mut _ as *mut _,
                size_of::<HMODULE>() as u32,
                &mut module_dyn_needed,
                filter_flag,
            )
        } {
            return Err("unable to EnumProcessModulesEx");
        }

        let mut name: Vec<WCHAR> = vec![0; winapi::shared::minwindef::MAX_PATH];
        // https://docs.microsoft.com/en-us/windows/win32/api/psapi/nf-psapi-getmodulebasenamew
        let len_name = unsafe {
            K32GetModuleBaseNameW(
                handle,
                module_ptr,
                name.as_mut_ptr(),
                (name.len() / size_of::<WCHAR>()) as u32,
            )
        };
        if len_name == 0 {
            return Err("unable to K32GetModuleBaseNameW");
        }

        let name = String::from_utf16_lossy(&name[0..len_name as usize]);

        Ok(ProcessHandle {
            pid: self,
            handle,
            name,
        })
    }
}
