//use winapi::um::winnt::SystemPowerInformation;
use std::mem::size_of; // buffer/struct size
use std::ptr::null_mut; // NULL
use std::time::Duration;

use winapi::shared::minwindef::{DWORD, HMODULE, PULONG, UCHAR, ULONG};
use winapi::um::errhandlingapi::GetLastError;
use winapi::um::handleapi::CloseHandle;
use winapi::um::processthreadsapi::{GetCurrentProcess, OpenProcess};
use winapi::um::psapi::{K32EnumProcessModulesEx, K32EnumProcesses, K32GetModuleBaseNameW};
use winapi::um::systemtopologyapi::GetNumaHighestNodeNumber;
use winapi::um::winbase::{GetProcessAffinityMask, SetProcessAffinityMask};
use winapi::um::winnt::{HANDLE, WCHAR};

/*
#[derive(Debug)]
pub struct ProcessHandle {
    pid: Pid,
    handle: HANDLE,
    name: String,
}
*/

impl Drop for ProcessHandle {
    fn drop(&mut self) {
        unsafe { CloseHandle(self.handle) };
    }
}

pub fn get_cpu_info() -> Result<CpuInfo, &'static str> {
    let mut numa_node_count: ULONG = 0xffffffff;
    // https://docs.microsoft.com/en-us/windows/win32/api/systemtopologyapi/nf-systemtopologyapi-getnumahighestnodenumber
    let r = unsafe { GetNumaHighestNodeNumber(&mut numa_node_count) };
    if r == 0 {
        println!("Error: {}", os_get_last_error());
        println!(
            "(DD) GetNumaHighestNodeNumber({}) -> {}",
            numa_node_count, r
        );
        return Err("unable to get GetNumaHighestNodeNumber");
    }

    if numa_node_count > 0 {
        //Ok(numa_node_count);
        unimplemented!("NUMA Systems not implemented yet!");
    } else {
        let proc_handle = unsafe { GetCurrentProcess() };
        let mut lpProcessAffinityMask: usize = 0b0010_0000;
        //unsafe { SetProcessAffinityMask(proc_handle, lpProcessAffinityMask as u32) };
        let mut lpSystemAffinityMask: usize = usize::max_value();
        let r = unsafe {
            GetProcessAffinityMask(
                proc_handle,
                &mut lpProcessAffinityMask as *mut _ as *mut _,
                &mut lpSystemAffinityMask as *mut _ as *mut _,
            )
        };
        if r == 0 {
            return Err("unable to get GetProcessAffinityMask()");
        }
        Ok(CpuInfo {
            all_system_cores: mask_to_affinity(lpSystemAffinityMask as u32),
        })
    }
}

fn affinity_to_mask(l: &Vec<CoreId>) -> u32 {
    let mut mask = 0;

    for i in l {
        mask |= 1u32 << i.0
    }

    mask
}
fn mask_to_affinity(mask: u32) -> Vec<CoreId> {
    let mut r = Vec::with_capacity(mask.count_ones() as usize);

    for i in 0..32 {
        if mask & (1 << i) > 0 {
            r.push(CoreId(i));
        }
    }

    r
}

pub fn set_process_affinity(name: &str, cores: &Vec<CoreId>, all_cores: &Vec<CoreId>) {
    let pids = Pid::get_system_pids();
    // println!("(DD) pids {:#?}", pids);

    let mut handles: Box<Iterator<Item = ProcessHandle>> = Box::new(
        pids.unwrap()
            .into_iter()
            .map(Pid::process_handle)
            .filter_map(Result::ok),
    );
    if name != "*" {
        handles = Box::new(handles.filter(|p| p.name == name));
    }
    //.collect::<Vec<_>>()
    //println!("(DD) handles {:#?}", handles);
    for ph in handles {
        let cores = if ph.name.starts_with("cargo") {
            all_cores
        } else {
            cores
        };
        let status = if 0 == unsafe { SetProcessAffinityMask(ph.handle, affinity_to_mask(cores)) } {
            os_get_last_error()
        } else {
            "Ok".into()
        };
        println!("{}: (PID: {}) <= {}", ph.name, ph.pid.0, status);
    }
}

fn os_get_last_error() -> String {
    use winapi::shared::ntdef::{LANG_NEUTRAL, LPWSTR, MAKELANGID, SUBLANG_DEFAULT};
    use winapi::um::winbase::*;
    let code = unsafe { GetLastError() };

    let mut str_ptr: LPWSTR = null_mut();
    let dword_len = unsafe {
        winapi::um::winbase::FormatMessageW(
            FORMAT_MESSAGE_ALLOCATE_BUFFER
                | FORMAT_MESSAGE_FROM_SYSTEM
                | FORMAT_MESSAGE_IGNORE_INSERTS,
            null_mut(),
            code,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) as u32,
            &mut str_ptr as *mut _ as *mut _,
            0,
            null_mut(),
        )
    };
    let msg_slice = unsafe { std::slice::from_raw_parts(str_ptr, dword_len as usize) };

    let r = format!(
        "Code: {}; Message: {}",
        code,
        String::from_utf16_lossy(msg_slice)
    );

    unsafe { winapi::um::winbase::LocalFree(str_ptr as *mut _) };

    r
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn vec_to_mask() {
        let expected = 0b0001_0101;
        let result = affinity_to_mask(&vec![CoreId(2), CoreId(4), CoreId(0)]);
        assert_eq!(expected, result);
    }
    #[test]
    fn mask_to_vec() {
        let expected = vec![CoreId(0), CoreId(2), CoreId(4)];
        let result = mask_to_affinity(0b0001_0101);
        assert_eq!(expected, result);
    }
}
